import os
from typing import Optional

import pandas as pd
from pydantic import validate_call
from syndb_api_client import ApiException, AuthApi, MetadataApi, NeuroanatomicalDirection

from syndb_data.app_config import ClientConfig
from syndb_data.validator import UserInfo, exception_from_bad_response


def cache_all_tags(client_config: Optional[ClientConfig] = None) -> None:
    client_config = client_config or ClientConfig()

    client_config.tag_cache_dir.mkdir(exist_ok=True)

    with client_config.get_api_client(requires_auth=True) as api_client:
        try:
            all_tags = MetadataApi(api_client).fetch_all_tags_endpoint_neurodata_metadata_tag_get_all_get()
        except ApiException as e:
            exception_from_bad_response(
                "MetadataApi",
                "fetch_all_tags_endpoint_neurodata_metadata_tag_get_all_get",
                e,
            )

    if all_tags.animal:
        pd.Series([a.species for a in all_tags.animal]).to_csv(client_config.animal_species_cache_file_path)

        animal_cultural = pd.Series([a.cultural_name for a in all_tags.animal])
        if animal_cultural.empty:
            animal_cultural.drop_duplicates(inplace=True)
            animal_cultural.to_csv(client_config.animal_cultural_name_cache_file_path)

    if all_tags.brain_region:
        brain_region_records = [
            b.brain_structure if b.direction == NeuroanatomicalDirection.NULL else f"{b.direction}-{b.brain_structure}"
            for b in all_tags.brain_region
        ]
        pd.Series(brain_region_records).to_csv(client_config.brain_region_cache_file_path)

    if all_tags.mutation:
        pd.Series([f"{m.gene}-{m.expression}" for m in all_tags.mutation]).to_csv(
            client_config.mutation_cache_file_path
        )

    if all_tags.microscopy_names:
        pd.Series(all_tags.microscopy_names).to_csv(client_config.microscopy_cache_file_path)

    if all_tags.publication_dois:
        pd.Series(all_tags.publication_dois).to_csv(client_config.publication_cache_file_path)


@validate_call
def register_user(
    email: str,
    password: str,
    orcid: bool = False,
    *,
    client_config: Optional[ClientConfig] = None,
) -> UserInfo:
    if orcid:
        raise NotImplementedError

    client_config = client_config or ClientConfig()
    assert not list(
        client_config.local_app_directory.iterdir()
    ), f"Cannot create a new user while a user is logged in, delete the app folder {client_config.local_app_directory}"

    with client_config.get_api_client() as api_client:
        api_instance = AuthApi(api_client)
        email = email or input("email: ").strip()
        try:
            api_response = api_instance.auth_database_login_user_auth_database_login_post(
                username=email, password=password
            )
        except ApiException as e:
            exception_from_bad_response("AuthApi", "auth_database_login_user_auth_database_login_post", e)

    return client_config.save_logged_in_user(email, api_response.access_token)


@validate_call
def logout_from_local_user(client_config: Optional[ClientConfig] = None) -> None:
    client_config = client_config or ClientConfig()

    with client_config.get_api_client(requires_auth=True) as api_client:
        try:
            api_response = AuthApi(api_client).auth_database_logout_user_auth_database_logout_post()
        except ApiException as e:
            exception_from_bad_response("AuthApi", "auth_database_login_user_auth_database_login_post", e)

    os.remove(client_config.user_config_path)

    assert api_response
