import csv
import json
from getpass import getpass
from pathlib import Path
from typing import ClassVar, Optional
from uuid import UUID

import rtoml
from pydantic import DirectoryPath, FilePath, computed_field, field_validator, validate_call
from pydantic_settings import BaseSettings
from syndb_api_client import ApiClient, ApiException, AuthApi, Configuration
from syndb_constants.app import TEST_PASSWORD, TEST_USERNAME

from syndb_data.validator import UserInfo, exception_from_bad_response


@validate_call
def _append_to_csv(csv_path: Path, row: list[str]) -> None:
    csv_path.touch(exist_ok=True)
    with open(csv_path, mode="a", newline="") as file:
        writer = csv.writer(file)
        writer.writerow(row)


class ClientConfig(BaseSettings):
    local_app_directory: DirectoryPath = Path.home() / ".syndb"
    cached_cave_auth_key_not_found_error_msg: str = (
        "CAVE authentication key has not been cached, define before retrieval"
    )

    testing: bool = False

    cache_format: ClassVar[str] = ".csv"
    cave_json_cache_key: ClassVar[str] = "cave_auth_key"

    _production_url: ClassVar[str] = "https://syndb.encods.eu/api/v1"
    _testing_url: ClassVar[str] = "http://localhost:8180"

    @field_validator("local_app_directory", mode="before")
    def ensure_cache_directory(cls, value):
        value = Path(value)
        value.mkdir(parents=True, exist_ok=True)
        return value

    @computed_field  # type: ignore[misc]
    @property
    def user_config_path(self) -> FilePath:
        return self.local_app_directory / "user_config.toml"

    @computed_field  # type: ignore[misc]
    @property
    def api_url(self) -> str:
        return self._testing_url if self.testing else self._production_url

    def get_api_client(self, requires_auth: bool = False) -> ApiClient:
        return ApiClient(
            Configuration(
                host=self.api_url,
                access_token=self.get_credentials().access_token if requires_auth else None,
            )
        )

    def get_credentials(
        self,
        email: Optional[str] = None,
        password: Optional[str] = None,
        renew: bool = False,
    ) -> UserInfo:
        if not renew and self.user_config_path.exists():
            return UserInfo(**rtoml.load(self.user_config_path))

        if self.testing:
            assert not email, "SYNDB_TESTING is only for testing purposes, please set to null in the env vars"
            email = TEST_USERNAME
            password = TEST_PASSWORD
        else:
            email = email or input("email: ")
            password = password or getpass("password: ")

        with self.get_api_client() as api_client:
            email = email or input("email: ").strip()
            try:
                api_response = AuthApi(api_client).auth_database_login_user_auth_database_login_post(
                    username=email, password=password
                )
            except ApiException as e:
                exception_from_bad_response("AuthApi", "auth_database_login_user_auth_database_login_post", e)

        return self.save_logged_in_user(email, api_response.access_token)

    def save_logged_in_user(self, email: str, access_token: str) -> UserInfo:
        result = UserInfo(email=email, access_token=access_token)
        rtoml.dump(result.model_dump(), self.user_config_path)
        return result

    # CACHE ==================

    @validate_call
    def parquet_out_dir_to_cache(self, parquet_out_dir_path: DirectoryPath) -> None:
        _append_to_csv(self.out_parquet_dir_cache_file_path, [str(parquet_out_dir_path)])

    @validate_call
    def dataset_to_cache(self, dataset_id: UUID, label: str) -> None:
        _append_to_csv(self.dataset_cache_file_path, [str(dataset_id), label])

    @computed_field  # type: ignore[misc]
    @property
    def cache_dir(self) -> DirectoryPath:
        return self.local_app_directory / "cache"

    @computed_field  # type: ignore[misc]
    @property
    def out_parquet_dir_cache_file_path(self) -> FilePath:
        return self.tag_cache_dir / f"out_parquet_dir{self.cache_format}"

    @computed_field  # type: ignore[misc]
    @property
    def dataset_cache_file_path(self) -> FilePath:
        return self.tag_cache_dir / f"out_parquet_dir{self.cache_format}"

    @computed_field  # type: ignore[misc]
    @property
    def tag_cache_dir(self) -> DirectoryPath:
        result = self.local_app_directory / "tags"
        result.mkdir(exist_ok=True)
        return result

    @computed_field  # type: ignore[misc]
    @property
    def animal_species_cache_file_path(self) -> FilePath:
        return self.tag_cache_dir / f"animal_species{self.cache_format}"

    @computed_field  # type: ignore[misc]
    @property
    def animal_cultural_name_cache_file_path(self) -> FilePath:
        return self.tag_cache_dir / f"animal_cultural_name{self.cache_format}"

    @computed_field  # type: ignore[misc]
    @property
    def brain_region_cache_file_path(self) -> FilePath:
        return self.tag_cache_dir / f"brain_region{self.cache_format}"

    @computed_field  # type: ignore[misc]
    @property
    def mutation_cache_file_path(self) -> FilePath:
        return self.tag_cache_dir / f"mutation{self.cache_format}"

    @computed_field  # type: ignore[misc]
    @property
    def microscopy_cache_file_path(self) -> FilePath:
        return self.tag_cache_dir / f"microscopy{self.cache_format}"

    @computed_field  # type: ignore[misc]
    @property
    def publication_cache_file_path(self) -> FilePath:
        return self.tag_cache_dir / f"publication{self.cache_format}"

    @computed_field  # type: ignore[misc]
    @property
    def cave_auth_key_cache_file_path(self) -> FilePath:
        return self.cache_dir / "cave_auth_key.json"

    def cache_cave_auth_key(self, cave_auth_key: str) -> None:
        with open(self.cave_auth_key_cache_file_path, "w") as out_json:
            json.dump({self.cave_json_cache_key: cave_auth_key}, out_json)

    def retrieve_cave_auth_key(self) -> str:
        with open(self.cave_auth_key_cache_file_path, "r") as in_json:
            return json.load(in_json)[self.cave_json_cache_key]
