import tempfile
from pathlib import Path
from uuid import UUID

import caveclient
import pandas as pd
from pydantic import validate_call
from syndb_constants.dataset import (
    ACETYLCHOLINE_COLUMN_NAME,
    CID_COLUMN_NAME,
    CLEFT_SCORE_COLUMN_NAME,
    CONNECTION_SCORE_COLUMN_NAME,
    DOPAMINE_COLUMN_NAME,
    GABA_COLUMN_NAME,
    GLUTAMATE_COLUMN_NAME,
    OCTOPAMINE_COLUMN_NAME,
    SERINE_COLUMN_NAME,
)
from uuid6 import uuid7

from syndb_data.dataset.io import upload_neurodata_parquets


@validate_call
def cave_datastack_syndb_transfer(cave_datastack_name: str, dataset_id: UUID, cave_auth_key: str | None = None) -> None:
    """
    Transfer a CAVE datastack to SynDB. The data will be stored under a single dataset, the dataset is defined by
    its ID.

    To use this function, you must have a CAVE authentication key. To get one:
    1) Go to: https://global.daf-apis.com/auth/api/v1/create_token to create a new token.
    2) Log in with your Google credentials and copy the token shown afterward.

    Either pass the token to this function, or you can save the token to the local computer:
        client.auth.save_token(token="PASTE_YOUR_TOKEN_HERE")

    :param cave_datastack_name: CAVE datastack name
    :param dataset_id: SynDB dataset ID
    :param cave_auth_key: Manually defined CAVE authentication key. Must have a local token saved if not defined.
    :return: None
    """
    client = caveclient.CAVEclient(cave_datastack_name, auth_token=cave_auth_key)

    syndb_data = []
    for row in client.materialize.query_view("valid_synapses_nt_np").itertuple(index=False):
        syndb_data.append(
            [
                uuid7().bytes,
                row["id"],
                row["connection_score"],
                row["cleft_score"],
                row["gaba"],
                row["ach"],
                row["glut"],
                row["oct"],
                row["ser"],
                row["da"],
            ]
        )

    df = pd.DataFrame(
        syndb_data,
        columns=[
            CID_COLUMN_NAME,
            CONNECTION_SCORE_COLUMN_NAME,
            CLEFT_SCORE_COLUMN_NAME,
            GABA_COLUMN_NAME,
            ACETYLCHOLINE_COLUMN_NAME,
            GLUTAMATE_COLUMN_NAME,
            OCTOPAMINE_COLUMN_NAME,
            SERINE_COLUMN_NAME,
            DOPAMINE_COLUMN_NAME,
        ],
    )

    with tempfile.TemporaryDirectory() as tmp_dir:
        data_path = Path(tmp_dir) / "SYNAPSE.parquet"
        df.to_parquet(data_path)
        upload_neurodata_parquets(dataset_id, [data_path])
