import asyncio
from typing import Mapping, Optional
from uuid import UUID

import pandas as pd
from pydantic import DirectoryPath, FilePath, validate_call
from syndb_api_client import ApiException, CreateDataset, DatasetApi, DatasetSearchModel, IoApi, NeurodataApi
from syndb_constants.dataset import (
    MESH_PATH_COLUMN_NAME,
    S3_MESH_LOCATION_COLUMN_NAME,
    S3_SWB_LOCATION_COLUMN_NAME,
    SWB_DF_COLUMN_NAME,
)
from syndb_constants.schema import model_to_schema
from syndb_constants.table import SyndbTable, syndb_table_name_to_enum, syndb_table_supported_str

from syndb_data.app_config import ClientConfig
from syndb_data.dataset.s3 import upload_neurodata_to_s3
from syndb_data.validator import exception_from_bad_response, validate_parquet_files


@validate_call
def create_dataset(dataset: CreateDataset, client_config: Optional[ClientConfig] = None) -> UUID:
    client_config = client_config or ClientConfig()

    with client_config.get_api_client(requires_auth=True) as api_client:
        try:
            api_response = NeurodataApi(api_client).register_new_dataset_neurodata_metadata_dataset_new_dataset_post(
                dataset
            )
        except ApiException as e:
            exception_from_bad_response(
                "NeurodataApi",
                "register_new_dataset_neurodata_metadata_dataset_register_new_post",
                e,
            )

    dataset_id = UUID(api_response.id)
    client_config.dataset_to_cache(dataset_id, dataset.label)

    return dataset_id


@validate_call
def upload_neurodata_parquets(
    dataset_id: UUID,
    parquet_paths: list[FilePath],
    client_config: Optional[ClientConfig] = None,
) -> None:
    client_config = client_config or ClientConfig()

    validate_parquet_files(parquet_paths)

    asyncio.run(upload_neurodata_to_s3(parquet_paths, dataset_id, client_config))

    with client_config.get_api_client(requires_auth=True) as api_client:
        try:
            IoApi(api_client).upload_dataset_files_neurodata_io_upload_dataset_files_post(
                str(dataset_id), [str(p) for p in parquet_paths]
            )
        except ApiException as e:
            exception_from_bad_response(
                "NeurodataApi", "upload_dataset_files_neurodata_io_upload_dataset_files_post", e
            )


@validate_call
def dataset_tag_search(
    dataset_search: DatasetSearchModel, client_config: Optional[ClientConfig] = None
) -> Mapping[UUID, str]:
    client_config = client_config or ClientConfig()

    with client_config.get_api_client(requires_auth=True) as api_client:
        try:
            api_response = DatasetApi(
                api_client
            ).dataset_tag_search_endpoint_neurodata_metadata_dataset_dataset_tag_search_post(dataset_search)
        except ApiException as e:
            exception_from_bad_response(
                "NeurodataApi",
                "dataset_tag_search_neurodata_metadata_dataset_dataset_tag_search_post",
                e,
            )

    return {UUID(uuid_str): label for uuid_str, label in api_response.uuid_to_label.items()}


@validate_call(config=dict(arbitrary_types_allowed=True))
def validate_dataset_dataframe(df: pd.DataFrame, syndb_table: SyndbTable) -> pd.DataFrame:
    """
    Validate dataset by comparing it to the Scylla DB schema

    :param df: pd.DataFrame that will be validated
    :param syndb_table: The enum representing the table the data frame belongs
    :return: Validated data frame
    """
    column_to_schema = set(model_to_schema[syndb_table.name.lower()])

    for server_side_only, client_side in (
        (S3_MESH_LOCATION_COLUMN_NAME, MESH_PATH_COLUMN_NAME),
        (S3_SWB_LOCATION_COLUMN_NAME, SWB_DF_COLUMN_NAME),
    ):
        if server_side_only in column_to_schema:
            column_to_schema.remove(server_side_only)
            column_to_schema.add(client_side)

    assert not (
        diff := frozenset(df.columns).difference(column_to_schema)
    ), f"Following columns have the wrong names: {diff}, choose from {column_to_schema}"
    return df


@validate_call
def read_syndb_table_data_from_directory(
    data_dir: DirectoryPath, read_dfs: bool = True
) -> dict[SyndbTable, pd.DataFrame] | dict[SyndbTable, FilePath]:
    """
    Read the SynDB data from files in the provided directory (data_dir)

    :param data_dir: The directory to scan for data, the file names must include compartment (SyndbTable) names
    :param read_dfs: When True, read the data in the path to a pd.DataFrame, suffix agnostic
    :return:
    """
    result = {}
    for df_file in data_dir.iterdir():
        if df_file.is_dir():
            continue

        found = False
        for name in syndb_table_name_to_enum:
            if name in df_file.stem.lower():
                assert not found, f"Found multiple table references in file name {df_file.stem}"

                result[syndb_table_name_to_enum[name]] = _read_dataframe(df_file) if read_dfs else df_file
                found = True

        if not found:
            raise ValueError(f"{df_file} must have a syndb_table enum as its name: {syndb_table_supported_str}")

    return result


@validate_call
def read_syndb_table_data_from_file_path(
    data_paths: list[FilePath], read_dfs: bool = True
) -> dict[SyndbTable, pd.DataFrame] | dict[SyndbTable, FilePath]:
    """
    Read the SynDB data from provided file paths (data_paths)

    :param data_paths: Paths to data files, the file names must include compartment (SyndbTable) names
    :param read_dfs: When True, read the data in the path to a pd.DataFrame, suffix agnostic
    :return:
    """
    result = {}
    for syndb_table_name, syndb_table in syndb_table_name_to_enum.items():
        for data_path in data_paths:
            if syndb_table_name in data_path.stem:
                assert syndb_table not in result
                result[syndb_table] = _read_dataframe(data_path) if read_dfs else data_path
    return result


@validate_call
def _read_dataframe(file_path: FilePath) -> pd.DataFrame:
    """
    File-type agnostic pandas file reader.

    :param file_path: The file to read
    :return: pd.DataFrame
    """
    match file_path.suffix:
        case ".xls" | ".xlsx" | ".xlsm" | ".xlsb" | ".odf" | ".ods" | ".odt":
            return pd.read_excel(file_path)
        case ".csv":
            return pd.read_csv(file_path)
        case ".json":
            return pd.read_json(file_path)
        case ".parquet":
            return pd.read_parquet(file_path)
        case ".feather":
            return pd.read_feather(file_path)
        case ".dta":
            return pd.read_stata(file_path)
        case ".sav":
            return pd.read_spss(file_path)
        case ".pkl" | ".pickle":
            return pd.read_pickle(file_path)
        case _:
            raise ValueError(f"{file_path.name} must have a syndb_table enum as its name: {syndb_table_supported_str}")
