import os
from pathlib import Path
from shutil import rmtree
from typing import Hashable, Mapping, Optional
from uuid import UUID

import pandas as pd
from pydantic import DirectoryPath, FilePath, validate_call
from syndb_constants.table import (
    ID_COLUMN_NAME,
    PARENT_ENUM_COLUMN_NAME,
    PARENT_ID_COLUMN_NAME,
    SyndbTable,
    compartment_to_valid_parents,
    syndb_table_hierarchy_descending,
    syndb_table_name_to_enum,
)
from uuid6 import uuid7

from syndb_data.app_config import ClientConfig
from syndb_data.dataset.io import validate_dataset_dataframe, read_syndb_table_data_from_directory, \
    read_syndb_table_data_from_file_path


def _replace_id_with_uuid_in_df(
    df: pd.DataFrame,
) -> tuple[pd.DataFrame, Mapping[Hashable, bytes]]:
    """
    Updates the input data frame where the ID column is replaced with UUID column

    :param df: Data frame that will have its ID column replaced
    :return: Data frame with updated ID column, map from old ID to new UUID
    """
    id_column = df[ID_COLUMN_NAME]
    old_to_new = {old: uuid7().bytes for old in frozenset(id_column)}
    df.loc[:, ID_COLUMN_NAME] = [old_to_new[old] for old in id_column]
    return df, old_to_new


def _create_id_hierarchy(
    syndb_table_to_df: dict[SyndbTable, pd.DataFrame]
) -> tuple[dict[SyndbTable, pd.DataFrame], Mapping[SyndbTable, Mapping[Hashable, UUID]]]:
    table_to_org_id_to_uuid: dict = {}
    for syndb_table in syndb_table_hierarchy_descending:
        if syndb_table not in syndb_table_to_df:
            continue

        (
            syndb_table_to_df[syndb_table],
            table_to_org_id_to_uuid[syndb_table],
        ) = _replace_id_with_uuid_in_df(syndb_table_to_df[syndb_table])

    return syndb_table_to_df, table_to_org_id_to_uuid


@validate_call(config=dict(arbitrary_types_allowed=True))
def simple_id_replace(df: pd.DataFrame, syndb_table: Optional[SyndbTable] = None) -> pd.DataFrame:
    """
    Replace ID column with UUIDs. Validates against SynDB schema if syndb_table is provided

    :param df: DataFrame that will have its ID column replaced with a UUID column
    :param syndb_table: The SyndbTable enum the data belongs to
    :return: DataFrame with UUID column
    """
    result = _replace_id_with_uuid_in_df(df)[0]
    if syndb_table:
        validate_dataset_dataframe(result, syndb_table)
    return result


@validate_call(config=dict(arbitrary_types_allowed=True))
def hierarchical_df_closest_parent(syndb_table_to_df: dict[SyndbTable, pd.DataFrame]) -> dict[SyndbTable, pd.DataFrame]:
    """
    Update the ID and optionally parent column of a hierarchy of syndb table data frames

    :param syndb_table_to_df: Mapping from SyndbTable enum to data frame; concatenate before use in case multiple data frames per SyndbTable enum
    :return: Mapping from SyndbTable enum to data frame with UUIDs in ID and parent columns
    """
    syndb_table_to_df, table_to_org_id_to_uuid = _create_id_hierarchy(syndb_table_to_df)

    parentless = []
    for syndb_table, df in syndb_table_to_df.items():
        parent_id_to_uuid = None
        parent_enum = None
        for possible_parent in compartment_to_valid_parents[syndb_table]:
            if possible_parent in table_to_org_id_to_uuid:
                parent_id_to_uuid = table_to_org_id_to_uuid[possible_parent]
                parent_enum = possible_parent
                break

        if not parent_id_to_uuid or not parent_enum:
            parentless.append(syndb_table)
            continue

        df[PARENT_ID_COLUMN_NAME] = [parent_id_to_uuid[old] for old in df[PARENT_ID_COLUMN_NAME]]
        df[PARENT_ENUM_COLUMN_NAME] = parent_enum.value

        validate_dataset_dataframe(df, syndb_table)

    if len(parentless) > 1:
        raise ValueError(f"{', '.join(str(p) for p in parentless)}: Lacks parent definitions")

    return syndb_table_to_df


@validate_call(config=dict(arbitrary_types_allowed=True))
def hierarchical_df_parent_designation(
    syndb_table_to_df: dict[SyndbTable, pd.DataFrame]
) -> dict[SyndbTable, pd.DataFrame]:
    """
    Update the ID and optionally parent column of a hierarchy of syndb table data frames

    Must define a column in the data frames to define the parent in a given row, see

    :param syndb_table_to_df: Mapping from SyndbTable enum to data frame; concatenate before use in case multiple data frames per SyndbTable enum
    :return: Mapping from SyndbTable enum to data frame with UUIDs in ID and parent columns
    """
    syndb_table_to_df, table_to_org_id_to_uuid = _create_id_hierarchy(syndb_table_to_df)

    for syndb_table, df in syndb_table_to_df.items():
        if PARENT_ID_COLUMN_NAME not in df.columns:
            continue

        new_id_column: list[UUID | None] = []
        parent_enum_column: list[int | None] = []
        for idx, row in enumerate(df.iter_rows(named=True)):
            if PARENT_ENUM_COLUMN_NAME in row:
                parent_enum = syndb_table_name_to_enum(row[PARENT_ENUM_COLUMN_NAME].strip().lower())
                new_id_column.append(table_to_org_id_to_uuid[parent_enum][row[ID_COLUMN_NAME]])
                parent_enum_column.append(parent_enum.value)
            else:
                new_id_column.append(None)
                parent_enum_column.append(None)

        df[ID_COLUMN_NAME] = new_id_column
        df[PARENT_ENUM_COLUMN_NAME] = parent_enum_column

        validate_dataset_dataframe(df, syndb_table)

    return syndb_table_to_df


def read_syndb_table_data(
    data_dir: Optional[DirectoryPath] = None,
    data_paths: Optional[list[FilePath]] = None,
    read_dfs: bool = True,
) -> dict[SyndbTable, pd.DataFrame] | dict[SyndbTable, FilePath]:
    assert not (data_dir and data_paths), "data_dir data_paths must be defined exclusively"

    if data_dir:
        return read_syndb_table_data_from_directory(data_dir, read_dfs)
    elif data_paths:
        return read_syndb_table_data_from_file_path(data_paths, read_dfs)

    raise ValueError("Either data_dir or data_paths must be defined")


@validate_call(config=dict(arbitrary_types_allowed=True))
def save_dataset(
    dataset: dict[SyndbTable, pd.DataFrame],
    out: Path,
    inspect: bool = True,
    client_config: Optional[ClientConfig] = None,
) -> dict[SyndbTable, DirectoryPath]:
    """
    Save processed dataset to the given directory. Will be validated before saved.

    :param dataset: The dataset map that will be saved
    :param out: The directory to save the data in; will be made if it doesn't exist
    :param inspect: When True, create excel versions of the data that can be inspected, recommended
    :return: None
    """
    if out.exists():
        assert out.is_dir(), "out must be a directory"
        assert not list(out.iterdir()), "out directory must be empty"
    else:
        os.makedirs(out)

    if inspect:
        inspect_out = out.with_name(f"{out.name}_inspect")
        rmtree(inspect_out, ignore_errors=True)
        inspect_out.mkdir()

    result = {}
    for syndb_table, df in dataset.items():
        validate_dataset_dataframe(df, syndb_table)

        save_path = out / f"{syndb_table.name}.parquet"
        df.to_parquet(save_path, compression="brotli")
        result[syndb_table] = save_path

        if inspect:
            df.to_excel(inspect_out / f"{syndb_table.name}.xlsx")

    if client_config := client_config or ClientConfig():
        client_config.parquet_out_dir_to_cache(out)

    return result
