import asyncio
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional
from uuid import UUID

import pandas as pd
from aiohttp import ClientSession, FormData
from pydantic import FilePath
from syndb_api_client import ApiException, IoApi, PreSignedS3Schema
from syndb_constants.dataset import (
    MESH_PATH_COLUMN_NAME,
    S3_MESH_LOCATION_COLUMN_NAME,
    S3_SWB_LOCATION_COLUMN_NAME,
    SWB_DF_COLUMN_NAME,
)
from syndb_constants.table import ID_COLUMN_NAME, syndb_table_name_to_enum

from syndb_data.app_config import ClientConfig
from syndb_data.validator import exception_from_bad_response


def _get_mesh_upload_url(dataset_id: UUID, client_config: ClientConfig) -> PreSignedS3Schema:
    with client_config.get_api_client(requires_auth=True) as api_client:
        try:
            api_response = IoApi(api_client).upload_dataset_mesh_neurodata_io_upload_mesh_get(str(dataset_id))
        except ApiException as e:
            exception_from_bad_response("IoApi", "upload_dataset_mesh_neurodata_io_upload_mesh_get", e)

    assert api_response
    return api_response


def _get_swb_upload_url(dataset_id: UUID, client_config: ClientConfig) -> PreSignedS3Schema:
    with client_config.get_api_client(requires_auth=True) as api_client:
        try:
            api_response = IoApi(api_client).upload_dataset_swb_neurodata_io_upload_swb_get(str(dataset_id))
        except ApiException as e:
            exception_from_bad_response("IoApi", "upload_dataset_swb_neurodata_io_upload_swb_get", e)

    assert api_response
    return api_response


@dataclass(frozen=True)
class S3Instance:
    client_file_path_column_name: str
    db_location_column_name: str
    file_suffix: str
    content_type: str
    upload_url_getter: Callable[[UUID, ClientConfig], PreSignedS3Schema]


s3_instances = [
    S3Instance(
        client_file_path_column_name=MESH_PATH_COLUMN_NAME,
        db_location_column_name=S3_MESH_LOCATION_COLUMN_NAME,
        file_suffix=".glb",
        content_type="model/gltf-binary",
        upload_url_getter=_get_mesh_upload_url,
    ),
    S3Instance(
        client_file_path_column_name=SWB_DF_COLUMN_NAME,
        db_location_column_name=S3_SWB_LOCATION_COLUMN_NAME,
        file_suffix=".swb",
        content_type="text/plain",
        upload_url_getter=_get_swb_upload_url,
    ),
]


async def load_file_and_s3_upload(
    session: ClientSession, generic_form_data: FormData, file_path: Path, upload_file_name: str, url: str
) -> None:
    data = deepcopy(generic_form_data)
    with file_path.open("rb") as my_mesh:
        data.add_field("file", my_mesh.read(), filename=upload_file_name)

    async with session.post(url, data=data) as response:
        if response.status != 200:
            raise ValueError(f"Upload failed with status: {response.status}, Details: {await response.text()}")


async def upload_neurodata_to_s3(
    parquet_paths: list[FilePath],
    dataset_id: UUID,
    client_config: Optional[ClientConfig] = None,
) -> pd.DataFrame | None:
    """
    Uploads the parquet files to S3. This function is meant to be used after the parquet files have been prepared,
    see `syndb_data.dataset.prepare`.

    :param parquet_paths: Paths to the parquet files
    :param dataset_id: The dataset ID to associate the uploads to
    :param client_config: SynDB client configuration
    :return:
    """
    client_config = client_config or ClientConfig()

    base_data = FormData()
    base_data.add_field("acl", "public-read")

    upload_routines = []
    async with ClientSession() as session:
        for df_file_path in parquet_paths:
            df = pd.read_parquet(df_file_path)

            df_has_s3_instance = False
            column_to_values = defaultdict(list)
            for s3_instance in s3_instances:
                if s3_instance.client_file_path_column_name not in df.columns:
                    continue

                df_has_s3_instance = True
                instance_form_data = deepcopy(base_data)

                pre_signed_s3 = s3_instance.upload_url_getter(dataset_id, client_config)
                for field, value in pre_signed_s3.fields.items():
                    instance_form_data.add_field(field, value)

                instance_form_data.add_field("Content-Type", s3_instance.content_type)

                instance_dataset_upload_url = f"{pre_signed_s3.upload_url}/{dataset_id}"
                for row in df.itertuples(index=False):
                    if s3_instance.client_file_path_column_name not in row:
                        column_to_values[s3_instance.db_location_column_name].append(False)
                        continue

                    file_data_id = row[ID_COLUMN_NAME]
                    column_to_values[s3_instance.db_location_column_name].append(True)

                    upload_routines.append(
                        load_file_and_s3_upload(
                            session,
                            instance_form_data,
                            Path(f"{instance_dataset_upload_url}/{file_data_id}"),
                            f"{file_data_id}{s3_instance.file_suffix}",
                            row[s3_instance.client_file_path_column_name],
                        )
                    )
                    if len(upload_routines) == 600:
                        await asyncio.gather(*upload_routines)
                        upload_routines = []

                df.drop(s3_instance.client_file_path_column_name, axis=1, inplace=True)

            if df_has_s3_instance:
                for column_name, values in column_to_values.items():
                    df[column_name] = values

                df.to_parquet(df_file_path, compression="brotli")

        if upload_routines:
            await asyncio.gather(*upload_routines)
